/*
    E2Key (TM) user mode driver for GNU/Linux
    Copyright (C) 2001 James Carter <james@jfc.org.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <stdlib.h>
#include <sys/io.h>
#include "e2key.h"

#define DATA Port
#define STATUS (Port+1)
#define CONTROL (Port+2)

#define ENABLE_WRITE 0x30
#define WRITE 0x40
#define READ 0x80

#define BUSY 0x80

uint16_t Port=0;

static void
delay(int l)
{
  for (;l;l--) {
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
    inb(0x80); inb(0x80); inb(0x80); inb(0x80); inb(0x80);
  }
}

static void
sendbits(e2data data, int bits)
{
  int i,b;

  for (i=0;i<bits;i++) {
    b=data&0x8000?2:0;
    data=data<<1;
    outb(0x5c|0xa0|b,DATA);
    delay(1);
    outb(0x5c|0xa0|b|0x1,DATA);
    delay(2);
    outb(0x5c|0xa0|b,DATA);
    delay(1);
  }
  return;
}

static void
command(e2addr command, e2addr addr)
{
  e2data d;

  sendbits(0x4000,2);
  d=(command|addr)<<8;
  sendbits(d,8);
  return;
}

e2err
e2key_port(uint16_t port)
{
  switch (port) {
    case 0x278:
    case 0x378:
    case 0x3bc:
      Port=port;
      ioperm(Port,3,1);
      ioperm(0x80,1,1);
      return e2OK;
    default:
      return e2PortAddressErr;
  }
}

e2err
e2key_read(e2addr addr, e2data *data)
{
  uint8_t r;
  e2data d=0;
  int bit;

  if (!Port) return e2NoPort;
  if (addr>63) return e2AddressErr;
  r=inb(STATUS)&BUSY;
  if (r) {
    outb(0x0c,DATA);
    outb(0xcd,CONTROL);
    delay(7);
    outb(0xcc,CONTROL);
    delay(7);
    outb(0xcd,CONTROL);
    delay(7);
    while (r) r=inb(STATUS)&BUSY;
  }
  outb(0x5c,DATA);
  command(READ,addr);
  outb(0x5c|0xa0|0x01,DATA);
  for (bit=16;bit;bit--) {
    delay(2);
    outb(0x5c|0xa0,DATA);
    d=(d<<1)|(inb(STATUS)&BUSY?1:0);
    delay(2);
    outb(0x5c|0xa0|0x01,DATA);
  }
  outb(0x5c|0xa0,DATA);
  delay(5);
  *data=d;
  return e2OK;
}

e2err
e2key_write(e2addr addr, e2data data)
{
  if (!Port) return e2NoPort;
  if (addr>63) return e2AddressErr;
  outb(0x5c,DATA);
  command(ENABLE_WRITE,0x0);
  outb(0x5c,DATA); outb(0x5c|0x1,DATA); outb(0x5c,DATA);
  delay(13);
  command(WRITE,addr);
  sendbits(data,16);
  outb(0x5c,DATA); outb(0x5c|0x1,DATA); outb(0x5c,DATA);
  delay(13);
  return e2OK;
}
